"""
Created on Mon Aug 25 13:17:03 2014

@author: anthony
"""
import time
from multiprocessing import Pool

import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate as interp

from .cp_tools import (
    cp_loglikelihood,
    cp_loglikelihood_proj,
    cp_model,
    mas2rad,
    project_cps,
    rad2mas,
)


def phase_binary_flux(u, v, wavel, p, return_cvis=False):
    """Calculate the phases observed by an array on a binary star
    ----------------------------------------------------------------
    p: 3-component vector (+2 optional), the binary "parameters":
    - p[0] = sep (mas)
    - p[1] = PA (deg) E of N.
    - p[2] = flux (primary is assumed to be 1)

    optional:
    - p[2:] = contrast ratio for several wavelengths that we want
             to calculate the cps over

    - u,v: baseline coordinates (meters)
    - wavel: wavelength (meters)
    ----------------------------------------------------------------"""
    p = np.array(p)
    # relative locations
    th = (p[1] + 90.0) * np.pi / 180.0
    ddec = mas2rad(p[0] * np.sin(th))
    dra = -mas2rad(p[0] * np.cos(th))

    # decompose into two "luminosities"
    # but first, a little trick so this works whether
    # p is a single value or a list of contrasts
    spec = p[2:]
    if len(spec) == 1:
        spec = spec[0]
    l2 = spec
    l1 = 1 - l2

    # phase-factor
    output_shape = list(u.shape)
    output_shape[-1] = np.size(wavel)
    phi = np.zeros(output_shape, dtype=complex)
    phi.real = np.cos(-2 * np.pi * (u * dra + v * ddec) / wavel)
    phi.imag = np.sin(-2 * np.pi * (u * dra + v * ddec) / wavel)

    cvis = l1 + l2 * phi

    phase = np.angle(cvis, deg=True)
    if return_cvis:
        return cvis
    else:
        return np.mod(phase + 10980.0, 360.0) - 180.0


# =========================================================================
def cp_model_flux(params, u, v, wavels, model="constant"):
    """Function to model closure phases. Takes a parameter list, u,v triangles and range of wavelengths.
    Allows fitting of a model to contrast vs wavelength.
    Models for contrast ratio:
            constant (contrast is constant with wavelength, default)
            linear (params[2,3]=contrast ratios at end wavelengths),
            free (params[2:]=contrast ratios).
            ndof (the wavelength channels are evenly spaced cubic interpolations in params[2:])
            polynomial (of the form Sum[n] params[n+2]*(wavelength*1e6)**n )
    NOTE: This doesn't allow for nonzero size of each component!"""
    nwav = wavels.size
    if model == "constant":
        cons = np.repeat(params[2], nwav)
    elif model == "linear":
        cons = params[2] + (params[3] - params[2]) * (wavels - wavels[0]) / (
            wavels[-1] - wavels[0]
        )
    elif model == "ndof":
        ndof = params[2:].size
        wavs = np.linspace(np.min(wavels), np.max(wavels), ndof)
        f = interp.interp1d(wavs, params[2:], kind="cubic")
        cons = f(wavels)
    elif model == "free":
        # no model, crat vs wav is free to vary.
        cons = params[2:]
    elif model == "polynomial":
        coefficients = params[2:]
        ndof = len(coefficients)
        cons = np.repeat(0.0, nwav)
        xax = (wavels - np.min(wavels)) / (np.max(wavels) - np.min(wavels))
        for order in range(ndof):
            cons += coefficients[order] * xax ** order
    else:
        raise NameError("Unknown model input to cp_model")

    # vectorize the arrays to speed up multi-wavelength calculations
    u = u[..., np.newaxis]  # (ncp x n_runs x 3 x 1) or (ncp x 3 x 1)
    v = v[..., np.newaxis]  # (ncp x n_runs x 3 x 1) or (ncp x 3 x 1)
    wavels = wavels[np.newaxis, np.newaxis, :]  # (1 x 1 x 1 x nwav) or (1x1xnwav)
    if u.ndim == 4:
        wavels = wavels[np.newaxis]
    phases = phase_binary_flux(u, v, wavels, params)
    cps = np.sum(phases, axis=-2)

    return cps


# =========================================================================
# =========================================================================


def cp_loglikelihood_proj_flux(
    params, u, v, wavel, proj_t3data, proj_t3err, proj, model="constant"
):
    """Calculate loglikelihood for projected closure phase data.
    Used both in the MultiNest and MCMC Hammer implementations.
    Here proj is the eigenvector array"""

    # hacky way to introduce priors
    #    if (params[2] > 50000) or (params[2] < 0.):
    #        return -np.inf
    if (params[0] > 350.0) or (params[0] < 0.0):
        return -np.inf
    if (params[1] > 360.0) or (params[1] < 0.0):
        return -np.inf

    cps = cp_model_flux(params, u, v, wavel, model=model)

    proj_mod_cps = project_cps(cps, proj)

    chi2 = np.sum(((proj_t3data - proj_mod_cps) / proj_t3err) ** 2)

    loglike = -chi2 / 2
    return loglike


# =========================================================================


def chi2_grid(everything):
    """Function for multiprocessing, does 2d chi2 grid for xy_grid"""
    cpo = everything["cpo"]
    chi2 = np.zeros((len(everything["ys"]), len(everything["cons"])))
    x = everything["x"]
    ys = everything["ys"]
    seps = np.sqrt(x ** 2 + ys ** 2)
    pas = np.angle(np.complex(0, 1) * ys + np.complex(1, 0) * x, True) % 360

    projected = everything["projected"]

    for ix in range(ys.size):
        for k, con in enumerate(everything["cons"]):
            params = [seps[ix], pas[ix], con]
            if projected:
                chi2[ix, k] = -2 * cp_loglikelihood_proj_flux(
                    params,
                    cpo.u,
                    cpo.v,
                    cpo.wavel,
                    cpo.proj_t3data,
                    cpo.proj_t3err,
                    cpo.proj,
                )
            else:
                chi2[ix, k] = -2 * cp_loglikelihood(
                    params, cpo.u, cpo.v, cpo.wavel, cpo.t3data, cpo.t3err
                )

    return chi2


# =========================================================================

# =========================================================================


def xy_grid(
    cpo,
    nxy=30,
    ncon=32,
    xymax="Default",
    cmin=10.0,
    cmax=500.0,
    threads=0,
    err_scale=1.0,
    extra_error=0.0,
    fix_crat=False,
    cmap="ds9cool",
    plot_as_mags=False,
    projected=False,
):

    """An attempt to copy Sylvestre's chi2 grid plots, using x and y instead
    of separation and position angle.

    Written by A Cheetham, with some parts stolen from other pysco/pymask routines."""

    # ------------------------
    # first, load your data!
    # ------------------------

    ndata = cpo.ndata

    u, v = cpo.u, cpo.v

    cpo.t3err = np.sqrt(cpo.t3err ** 2 + extra_error ** 2)
    cpo.t3err *= err_scale

    wavel = cpo.wavel

    w = np.array(np.sqrt(u ** 2 + v ** 2)) / np.median(wavel)

    if xymax == "Default":
        #        xymax = cpt.rad2mas(1./np.min(w/np.max(wavel)))
        xymax = rad2mas(1.0 / np.min(w))

    # ------------------------
    # initialise grid params
    # ------------------------

    xys = np.linspace(-xymax, xymax, nxy)
    #    cons = cmin  + (cmax-cmin)  * np.linspace(0,1,ncon)
    cons = np.linspace(cmin, cmax, ncon)

    if fix_crat != False:
        cons = np.array([fix_crat])
        ncon = 1

    # ------------------------
    # Calculate chi2 at each point
    # ------------------------

    tic = time.time()  # start the clock
    chi2 = np.zeros((nxy, nxy, ncon))
    if threads == 0:
        toc = time.time()
        for ix, x in enumerate(xys):
            everything = {
                "x": x,
                "cons": cons,
                "ys": xys,
                "cpo": cpo,
                "ix": ix,
                "projected": projected,
            }
            chi2[ix, :, :] = chi2_grid(everything)
            if (ix % 50) == 0:
                tc = time.time()
                print("Done " + str(ix) + ". Time taken: " + str(tc - toc) + "seconds")
                toc = tc
    else:
        all_vars = []
        for ix in range(nxy):
            everything = {
                "x": xys[ix],
                "cons": cons,
                "ys": xys,
                "cpo": cpo,
                "ix": ix,
                "projected": projected,
            }
            all_vars.append(everything)
        pool = Pool(processes=threads)
        chi2 = pool.map(chi2_grid, all_vars)
    tf = time.time()
    if tf - tic > 60:
        print("Total time elapsed: " + str((tf - tic) / 60.0) + "mins")
    elif tf - tic <= 60:
        print("Total time elapsed: " + str(tf - tic) + " seconds")
    chi2 = np.array(chi2)
    best_ix = np.where(chi2 == np.amin(chi2))

    # hack: if the best chi2 is at more than one location, take the first.
    bestx = xys[best_ix[0][0]]
    besty = xys[best_ix[1][0]]
    sep = np.sqrt(bestx ** 2 + besty ** 2)
    pa = np.angle(np.complex(bestx, besty), True) % 360
    best_params = [sep, pa, cons[best_ix[2][0]]]
    best_params = np.array(np.array(best_params).ravel())
    print("Separation " + str(best_params[0]) + " mas")
    print("Position angle " + str(best_params[1]) + " deg")
    print("Contrast Ratio " + str(best_params[2]))
    # ---------------------------------------------------------------
    #                        sum over each variable so we can visualise it all
    # ---------------------------------------------------------------
    temp_chi2 = ndata * chi2 / np.amin(chi2)
    like = np.exp(-(temp_chi2 - ndata) / 2)
    x_y = np.sum(like, axis=2)

    # ---------------------------------------------------------------
    #                        contour plot!
    # ---------------------------------------------------------------
    names = ["Chi2", "Likelihood", "Best Contrast Ratio"]
    plots = [np.min(chi2, axis=2), x_y, cons[np.argmin(chi2, axis=2)]]
    for ix, n in enumerate(names):

        plt.figure(n)
        plt.clf()
        # Plot it with RA on the X axis
        plt.imshow(
            plots[ix],
            extent=[np.amin(xys), np.amax(xys), np.amin(xys), np.amax(xys)],
            aspect="auto",
            cmap=cmap,
        )
        plt.colorbar()
        plt.ylabel("Dec (mas)")
        plt.xlabel("RA (mas)")

        plt.plot([0], [0], "wo")
        plt.xlim(xys[-1], xys[0])
        plt.ylim(xys[0], xys[-1])

    # ---------------------------------------------------------------
    #               And the detection limits that come for free!
    # ---------------------------------------------------------------
    chi2_null = np.sum((cpo.t3data / cpo.t3err) ** 2)
    # Define the detec limits to be the contrast at which chi2_binary - chi2_null < 25
    detecs = (chi2 - chi2_null) < 25
    detec_lim = np.zeros((nxy, nxy))
    for x_ix in range(nxy):
        for y_ix in range(nxy):
            detectable_cons = cons[detecs[x_ix, y_ix, :]]
            if len(detectable_cons) == 0:
                detec_lim[x_ix, y_ix] = cons[-1]
            else:
                detec_lim[x_ix, y_ix] = np.min(detectable_cons)

    if plot_as_mags:
        detec_lim_plot = -2.5 * np.log10(detec_lim)
    else:
        detec_lim_plot = detec_lim

    plt.figure(1)
    plt.clf()
    #    plt.imshow(detec_lim,extent=(xys[0],xys[-1],xys[0],xys[-1]),cmap=cmap)
    # Plot it with RA on the X axis
    plt.imshow(detec_lim_plot, extent=(xys[0], xys[-1], xys[0], xys[-1]), cmap=cmap)
    plt.colorbar()
    plt.title("Detection limits")
    plt.xlabel("RA (mas)")
    plt.ylabel("Dec (mas)")
    plt.xlim(xys[-1], xys[0])
    plt.ylim(xys[0], xys[-1])

    # And we should also print whether the likelihood peak is a detection
    #  according to the limits we just calculated
    limit_at_pos = detec_lim[best_ix[0][0], best_ix[1][0]]
    print("Contrast limit at best fit position: " + str(limit_at_pos))
    if limit_at_pos > best_params[2]:
        print("Detection!")
    else:
        print("No significant detection found")

    data = {
        "chi2": chi2,
        "like": like,
        "xys": xys,
        "cons": cons,
        "best_params": best_params,
        "detec_lim": detec_lim,
    }
    return data


# =========================================================================



import glob
import logging

import matplotlib.pyplot as plt
import numpy as np
import scipy
from scipy.io.idl import readsav
from termcolor import cprint

# import pymask.oifits
from . import oifits
from .cp_tools import project_cps

"""------------------------------------------------------------------------
cpo.py - Python class for manipulating oifits format closure phase data.
------------------------------------------------------------------------"""


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    green = "\x1b[32;5m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    )

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: green + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


# create logger with 'spam_application'
log = logging.getLogger("PYMASK")
log.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

ch.setFormatter(CustomFormatter())

log.addHandler(ch)
# log.debug("A quirky message only developers care about")
# log.info("Curious users might want to know this")
# log.warn("Something is wrong and any user should be informed")
# log.error("Serious stuff, this is red for a reason")
# log.critical("OH NO everything is on fire")


class cpo:
    """Class used to manipulate multiple closure phase datasets"""

    def __init__(self, oifits):
        # Default instantiation.
        if type(oifits) == str:
            oifits = [oifits]
        if isinstance(oifits, str):
            self.extract_from_oifits(oifits)
        elif isinstance(oifits, list):
            self.extract_multi_oifits(oifits)
        else:
            log.error("inputs must be an oifits file or a list of oifits files.")

    def extract_from_oifits(self, filename):
        """Extract closure phase data from an oifits file."""

        data = oifits.open(filename)
        self.name = ""

        self.ndata = len(data.t3)

        for j in data.wavelength:
            wavel = data.wavelength[j].eff_wave
            self.wavel = wavel

        self.nwavs = len(self.wavel)

        self.target = data.target[0].target

        t3data = []
        t3err = []
        self.u = np.zeros((self.ndata, 3))
        self.v = np.zeros((self.ndata, 3))

        for j, t3 in enumerate(data.t3):
            t3data.append(t3.t3phi)
            t3err.append(t3.t3phierr)
            self.u[j, :] = [t3.u1coord, t3.u2coord, -(t3.u1coord + t3.u2coord)]
            self.v[j, :] = [t3.v1coord, t3.v2coord, -(t3.v1coord + t3.v2coord)]

        self.t3data = np.array(t3data)
        self.t3err = np.array(t3err)

        # Also load the v2
        nv2 = len(data.vis2)
        vis2data = []
        vis2err = []
        self.v2_u = np.zeros(nv2)
        self.v2_v = np.zeros(nv2)
        for j, v2 in enumerate(data.vis2):
            vis2data.append(v2.vis2data)
            vis2err.append(v2.vis2err)
            self.v2_u[j] = v2.ucoord
            self.v2_v[j] = v2.vcoord
        self.vis2data = np.array(vis2data)
        self.vis2err = np.array(vis2err)

    def extract_multi_oifits(self, lfiles, verbose=True):
        """Extract closure phase data from a list of oifits files."""
        U, V, t3data, t3err = [], [], [], []
        for f in lfiles:
            self.extract_from_oifits(f)
            U.extend(self.u)
            V.extend(self.v)
            t3data.extend(self.t3data)
            t3err.extend(self.t3err)
            nbl = len(self.u)
        self.u = np.array(U)
        self.v = np.array(V)
        self.t3data = np.array(t3data)
        self.t3err = np.array(t3err)
        if verbose:
            cprint(
                r"PYMASK - %i oifits loaded : n=%ix%i=%i closure phases."
                % (len(lfiles), len(lfiles), nbl, len(t3data)),
                "green",
            )


class icpo:
    """Class used to manipulate multiple closure phase datasets.
    This one structures the data differently, so it can be used with a covariance
    matrix, or with projected closure phases"""

    def __init__(self, oifits=None, directory=None, tsize_targ=None, tsize_cal=None):
        # Default instantiation.

        # if the file is a complete (kpi + kpd) structure
        # additional data can be loaded.
        if oifits:
            try:
                self.extract_from_oifits(oifits)
            except Exception:
                print("Invalid file.")
        else:
            # Otherwise load directory from the IDL bs files
            self.extract_from_idl_directory(
                directory, tsize_targ=tsize_targ, tsize_cal=tsize_cal
            )

    def extract_from_oifits(self, filename):
        """Extract closure phase data from an oifits file."""

        data = oifits.open(filename)
        self.name = ""

        self.ndata = len(data.t3)

        for j in data.wavelength:
            wavel = data.wavelength[j].eff_wave
            self.wavel = wavel
            break
        self.nwavs = len(self.wavel)

        self.target = data.target[0].target

        t3data = []
        t3err = []
        self.u = np.zeros((self.ndata, 3))
        self.v = np.zeros((self.ndata, 3))

        for j, t3 in enumerate(data.t3):
            t3data.append(t3.t3phi)
            t3err.append(t3.t3phierr)
            self.u[j, :] = [t3.u1coord, t3.u2coord, -(t3.u1coord + t3.u2coord)]
            self.v[j, :] = [t3.v1coord, t3.v2coord, -(t3.v1coord + t3.v2coord)]

        self.t3data = np.array(t3data)
        self.t3err = np.array(t3err)

        # Also load the v2
        nv2 = len(data.vis2)
        vis2data = []
        vis2err = []
        self.v2_u = np.zeros(nv2)
        self.v2_v = np.zeros(nv2)
        for j, v2 in enumerate(data.vis2):
            vis2data.append(v2.vis2data)
            vis2err.append(v2.vis2err)
            self.v2_u[j] = v2.ucoord
            self.v2_v[j] = v2.vcoord
        self.vis2data = np.array(vis2data)
        self.vis2err = np.array(vis2err)

        print("  This probably doesnt work...")

    #########

    def extract_from_idl_directory(
        self,
        analysis_dir,
        tsize_targ=None,
        tsize_cal=None,
        idl_masking_dir="/Users/cheetham/code/masking/",
    ):
        """Import the closure phase data from the IDL pipeline directly
        from the bispectrum (bs) files"""
        # Get all of the bispectrum files
        bs_files = glob.glob(analysis_dir + "bs*.idlvar")
        bs_files.sort()

        # Load the cubeinfo file
        cubeinfo_file = glob.glob(analysis_dir + "cubeinfo*.idlvar")
        cubeinfo = readsav(cubeinfo_file[0])

        tsize = cubeinfo["olog"]["tsize"][0]
        uflip = cubeinfo["olog"]["uflip"][0]
        pa = cubeinfo["olog"]["pa"][0]
        del_pa = cubeinfo["olog"]["del_pa"][0]

        # Check that tsize is the right size (i.e. that we found all the files)
        assert len(bs_files) == len(tsize)

        # Find which ones are the targets and calibrators
        # Default is that targets are all objects with tsize < 0
        if tsize_targ is None:
            targ_ix = tsize < 0
        else:
            targ_ix = np.zeros(len(tsize), dtype=bool)
            for ix in tsize_targ:
                targ_ix[tsize == ix] = True

        # Default is that cals are all objects with tsize > 0
        if tsize_cal is None:
            cal_ix = tsize > 0
        else:
            cal_ix = np.zeros(len(tsize), dtype=bool)
            for ix in tsize_cal:
                cal_ix[tsize == ix] = True

        # Set up all the arrays
        # CPs
        n_data_per_obs = []
        cal_cps = []
        targ_cps = []
        print(f"Found {len(bs_files):4d} bs files from IDL")

        # Loop through and load them
        for ix, f in enumerate(bs_files):

            data = readsav(f)
            cps = np.angle(data["bs_all"], deg=True)
            mean_cp = data["cp"]
            err_cp = data["cp_sig"]
            # Add a wavelength dimension to non-IFU data
            if mean_cp.ndim == 1:
                cps = cps[np.newaxis, :]
                mean_cp = mean_cp[np.newaxis, :]
                err_cp = err_cp[np.newaxis, :]

            # Check if targ or cal, then add the cps there
            if cal_ix[ix]:
                if len(cal_cps) == 0:
                    cal_cps = cps
                    mean_cal_cps = mean_cp[:, :, np.newaxis]
                    std_cal_cps = err_cp[:, :, np.newaxis]
                else:
                    cal_cps = np.append(cal_cps, cps, axis=2)
                    mean_cal_cps = np.append(
                        mean_cal_cps, mean_cp[:, :, np.newaxis], axis=2
                    )
                    std_cal_cps = np.append(
                        std_cal_cps, err_cp[:, :, np.newaxis], axis=2
                    )
            elif targ_ix[ix]:

                if len(targ_cps) == 0:
                    # We have lots of setup to do for the first file, and then we'll reuse it for the rest
                    # Load the matched filter file since it contains some useful info
                    mf_file = idl_masking_dir + "templates/" + data["mf_file"]
                    mf_data = readsav(mf_file)

                    # Work out the wavelengths now
                    # u_ideal is the baseline divided by lambda so we can get them back from the first baseline like this
                    xy_coords = mf_data["xy_coords"]
                    bl2h_ix = mf_data["bl2h_ix"]
                    if "u_ideal" in mf_data.keys():  # For the closing triangle approach
                        wavs = (
                            float(xy_coords[0, bl2h_ix[0, 0]])
                            - xy_coords[0, bl2h_ix[0, 1]]
                        ) / mf_data["u_ideal"][..., 0]
                    else:  # For the matched filter approach
                        wavs = (
                            float(xy_coords[0, bl2h_ix[0, 0]])
                            - xy_coords[0, bl2h_ix[0, 1]]
                        ) / mf_data["u"][..., 0]

                    wavs = np.atleast_1d(wavs)

                    # Get the uv coords from the matched filter file
                    # This should be the same at all wavelengths
                    if data["u"].ndim == 1:
                        # in metres
                        u_coords = data["u"][mf_data["bs2bl_ix"]] * wavs[0]
                        # in metres
                        v_coords = data["v"][mf_data["bs2bl_ix"]] * wavs[0]
                    else:
                        # in metres
                        u_coords = data["u"][0, mf_data["bs2bl_ix"]] * wavs[0]
                        # in metres
                        v_coords = data["v"][0, mf_data["bs2bl_ix"]] * wavs[0]

                # Rotate the uv coords
                pa_rad = (pa[ix] - 0.5 * del_pa[ix]) * np.pi / 180.0
                u_coords1 = uflip * u_coords * np.cos(pa_rad) + v_coords * np.sin(
                    pa_rad
                )
                v_coords1 = -uflip * u_coords * np.sin(pa_rad) + v_coords * np.cos(
                    pa_rad
                )

                if len(targ_cps) == 0:
                    # keep this a list to make it easier to separate them
                    targ_cps = [cps.transpose((1, 2, 0))]
                    mean_targ_cps = mean_cp[:, :, np.newaxis]
                    std_targ_cps = err_cp[:, :, np.newaxis]
                    targ_u_coords = u_coords1[:, np.newaxis, :]
                    targ_v_coords = v_coords1[:, np.newaxis, :]
                else:
                    #                    targ_cps = np.append(targ_cps,cps,axis=2)
                    targ_cps.append(cps.transpose((1, 2, 0)))
                    mean_targ_cps = np.append(
                        mean_targ_cps, mean_cp[:, :, np.newaxis], axis=2
                    )
                    std_targ_cps = np.append(
                        std_targ_cps, err_cp[:, :, np.newaxis], axis=2
                    )
                    targ_u_coords = np.append(
                        targ_u_coords, u_coords1[:, np.newaxis, :], axis=1
                    )
                    targ_v_coords = np.append(
                        targ_v_coords, v_coords1[:, np.newaxis, :], axis=1
                    )

                n_data_per_obs.append(cps.shape[2])

        # Convert to the right units
        mean_targ_cps *= 180.0 / np.pi
        mean_cal_cps *= 180.0 / np.pi
        std_targ_cps *= 180.0 / np.pi
        std_cal_cps *= 180.0 / np.pi

        cal_cov = np.zeros((cal_cps.shape[0], cal_cps.shape[1], cal_cps.shape[1]))
        cal_cov_inv = 0 * cal_cov
        for wav_ix, wav_clps in enumerate(cal_cps):
            # Get the covariance and its inverse
            wav_cov = np.cov(wav_clps)
            wav_cov_inv = np.linalg.inv(wav_cov)

            cal_cov[wav_ix] = wav_cov
            cal_cov_inv[wav_ix] = wav_cov_inv

        # Bug fix:
        # The uv coordinates above are all positive. The third baseline should be flipped
        targ_u_coords[:, :, 2] *= -1
        targ_v_coords[:, :, 2] *= -1

        # Save everything to the cpo object (if given)
        self.cal_cps = np.transpose(cal_cps, axes=(1, 2, 0))
        #        self.targ_cps = np.transpose(targ_cps,axes=(1,2,0))
        self.targ_cps = targ_cps
        self.cal_cp_sig = np.transpose(std_cal_cps, axes=(1, 2, 0))
        self.targ_cp_sig = np.transpose(std_targ_cps, axes=(1, 2, 0))
        self.t3data = np.transpose(mean_targ_cps, axes=(1, 2, 0))
        self.t3err = np.transpose(std_targ_cps, axes=(1, 2, 0))
        self.u = targ_u_coords
        self.v = targ_v_coords
        self.wavel = wavs

        # ICPO specific info
        self.cal_cov = cal_cov
        self.cal_cov_inv = cal_cov_inv

        # Some useful info
        self.ncp = self.t3data.shape[0]
        self.nobs = self.t3data.shape[1]
        self.n_runs = self.nobs
        self.ndata = self.ncp * self.nobs
        self.nwavs = self.wavel.shape[0]
        self.name = ""

    def make_proj(
        self,
        cov_matrix,
        plot=True,
        n_remove=2,
        n_significant=None,
        tol=1e-2,
        calibrate=True,
        silent=False,
    ):
        """Make a projection matrix and project the raw closure phases onto it,
        using the covariance matrix given.
        Options:
            cov_matrix: Should be the covariance matrix that you want to diagonalise
                    during the projection
            n_significant: The number of eigenvalues to take. Since n_clp > n_indep_clp,
                    some eigenvalues will be discarded. This allows a set
                    number to be taken, rather than relying on an automatic
                    detection using "tol" (see below)
            tol: the minimum value used to detect significant eigenvalues. This is
                    ignored if n_significant is set
        """
        proj = []

        for wav_ix, wav in enumerate(self.wavel):
            if not silent:
                print(" Wavelength: " + str(wav_ix))
            #
            cov = cov_matrix[wav_ix, :, :]

            evals, evects = scipy.linalg.eigh(cov)

            # Put them in increasing order
            evects = evects[:, ::-1]
            evals = evals[::-1]

            if n_significant:
                good_evals = np.arange(len(evals)) < n_significant
                if not silent:
                    print("  Taking " + str(n_significant) + " eigenvalues")
            else:
                good_evals = evals > tol
                if not silent:
                    print(
                        "  Found "
                        + str(np.sum(good_evals))
                        + " eigenvalues above tolerance"
                    )

            if plot:
                plt.clf()
                plt.plot(evals)
                plt.xlabel("Mode number")
                plt.ylabel("Eigenvalue")
                plt.yscale("log")
                xl = plt.xlim()
                plt.plot([xl[0], xl[1]], [tol, tol], "--", label="Minimum eigenvalue")
                plt.xlim(xl)

            # Only (n-1)(n-2)/2 are independent
            # This is clearly shown in the evalues, with a clear break between 0-15 and >15
            evects = evects[:, good_evals]
            evals = evals[good_evals]

            # Throw away the ones that need calibrating
            evects = evects[:, n_remove:]
            evals = evals[n_remove:]
            if not silent:
                print("  Discarding first " + str(n_remove) + " modes")

            #        for ix in range(evals.size):
            #            evects[:,ix] /= np.sqrt(np.sum(evects[:,ix]**2)) # Make sure they have length 1 (i.e. normal)

            proj.append(evects)
        proj = np.array(proj)
        self.proj = proj
        # this assumes its the same number for all data...
        self.n_good = np.sum(good_evals) - n_remove

        print("  Projecting target data onto statistically indep. basis set")
        # Now project the data, starting with the raw data
        proj_t3data = np.zeros((self.n_good, self.nobs, self.nwavs))
        proj_t3err = np.zeros((self.n_good, self.nobs, self.nwavs))
        all_proj_cps = []
        for obs_ix in range(self.nobs):

            obs_cps = self.targ_cps[obs_ix]

            if calibrate:
                # Calibrate the target clps by subtracting the median calibrator cp
                systematic_cps = np.median(self.cal_cps, axis=1)
                obs_cps -= systematic_cps[:, np.newaxis, :]

            proj_cps = project_cps(obs_cps, proj)

            proj_t3data[:, obs_ix, :] = np.nanmean(proj_cps, axis=1)
            proj_t3err[:, obs_ix, :] = np.nanstd(proj_cps, axis=1) / np.sqrt(
                proj_cps.shape[1]
            )  # SEM

            all_proj_cps.append(proj_cps)

        self.proj_t3data = proj_t3data
        self.proj_t3err = proj_t3err
        return all_proj_cps


# =========================================================================


class PolDiffObs:
    """Class used to manipulate polarimetric differential observables"""

    def __init__(self, filename):
        # Default instantiation.

        try:
            self.extract_from_idlvar(filename)
        except Exception:
            print("Invalid files.")

    def extract_from_idlvar(self, filename):
        """Extract the differential visibility and closure phase data from
        the files output by the IDL masking pipeline."""

        # Use scipy's idl reading function to get a dictionary of all the saved variables
        data = readsav(filename)

        # The differential visibility quantities
        self.u = data["u"]
        self.v = data["v"]
        self.vis = data["vis"]
        self.vis_err = data["vis_err"]

        # Closure phase quantities
        self.bs_u = data["bs_u"]
        self.bs_v = data["bs_v"]
        self.cp = data["cp"]
        self.cp_err = data["cp_err"]

        # Differential phase quantities
        self.ph = data["ph"]
        self.ph_err = data["ph_err"]

        # Generic quantities (that arent implemented yet)


#        self.wavel=data['wavel']



"""
@author: Anthony Soulain (University of Sydney)

-------------------------------------------------------------------------
AMICAL: Aperture Masking Interferometry Calibration and Analysis Library
-------------------------------------------------------------------------

Set of functions to calibrate NRM data using a calibrator star data.

-------------------------------------------------------------------------
"""

import numpy as np
from astropy.io import fits
from astropy.stats import sigma_clip
from matplotlib import pyplot as plt
from munch import munchify as dict2class

from amical.dpfit import leastsqFit
from amical.tools import wtmn


def _v2varfunc(X, parms):
    """
    This is a function to be fed to leastfit for fitting to normv2var
    to a 5-parameter windshake-seeing function.
    """
    b_lengths = X[0]
    b_angles = X[1]

    a = parms["a"]
    b = parms["b"]
    c = parms["c"]
    d = parms["d"]
    e = parms["e"]

    Y = (
        a
        + b * b_lengths
        + c * b_lengths ** 2
        + d * b_lengths * abs(np.sin(b_angles))
        + e * b_lengths ** 2 * np.sin(b_angles) ** 2
    )

    return Y


def _apply_sig_clip(
    data, e_data, sig_thres=2, ymin=0, ymax=1.2, var="V2", display=False
):
    """Apply the sigma-clipping on the dataset and plot some diagnostic plots."""
    filtered_data = sigma_clip(data, sigma=sig_thres, axis=0)

    n_files = data.shape[0]
    n_pts = data.shape[1]  # baselines or bs number

    mn_data_clip, std_data_clip = [], []
    for i in range(n_pts):
        cond = filtered_data[:, i].mask
        data_clip = data[:, i][~cond]
        e_data_clip = e_data[:, i][~cond]
        cmn, std = wtmn(data_clip, weights=e_data_clip)
        mn_data_clip.append(cmn)
        std_data_clip.append(std)

    data_med = np.median(data, axis=0)

    if var == "V2":
        ylabel = r"Raw V$^2$"
        xlabel = "# baselines"
    else:
        ylabel = r"Raw closure Phases $\Phi$ [deg]"
        xlabel = "# bispectrum"
        ymin = -ymax

    if display:
        plt.figure()
        plt.title("CALIBRATOR")
        if n_files != 1:
            plt.plot(data[0], color="grey", alpha=0.2, label="Data vs. files")
            plt.plot(data[1:, :].T, color="grey", alpha=0.2)
        plt.plot(data_med, "g--", lw=1, label="Median value")
        for i in range(n_files):
            data_ind = data[i, :]
            cond = filtered_data[i, :].mask
            x = np.arange(len(data_ind))
            bad = data_ind[cond]
            x_bad = x[cond]
            if i == 0:
                plt.plot(
                    x_bad,
                    bad,
                    "rx",
                    ms=3,
                    label=r"Rejected points (>%i$\sigma$)" % sig_thres,
                )
            else:
                plt.plot(x_bad, bad, "rx", ms=3)
        plt.legend(loc="best", fontsize=9)
        plt.ylim(ymin, ymax)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.tight_layout()
    return np.array(mn_data_clip), np.array(std_data_clip)


def _calc_correction_atm_vis2(
    data, corr_const=1, nf=100, display=False, verbose=False, normalizedUncer=False
):
    """
    This function corrects V^2 for seeing and windshake. Use this on source and
    cal before division. Returns a multiplicative correction to the V^2.

    Parameters:
    -----------
    `data` {class}:
        class like containting results from extract_bs function,
    `corr_const` {float}:
        Correction constant (0.4 is good for V for NIRC experiment).
    Returns:
    --------
    `correction` {array}:
        Correction factor to apply.

    """

    vis = data.vis2
    avar = data.matrix.avar
    u = data.u
    v = data.v

    err_avar = data.matrix.err_avar
    err_vis = data.e_vis2

    w = np.where((vis == 0) | (avar == 0))

    if len(w) == 0:
        print("Cannot continue - divide by zero imminent...")
        correction = None

    normvar = avar / vis ** 2
    w0 = np.where(u ** 2 + v ** 2 >= 0)

    param = {"a": 0, "b": 0, "c": 0, "d": 0, "e": 0}

    if len(err_avar) == 0:
        err_avar = normvar * 2.0 / np.sqrt(nf)
    if len(err_vis) != 0:
        err_avar = abs(normvar) * np.sqrt(
            err_avar ** 2 / avar ** 2 + 2.0 * err_vis ** 2 / vis ** 2
        )

    X = [u[w0], v[w0]]
    fit = leastsqFit(
        _v2varfunc,
        X,
        param,
        normvar[w0],
        err_avar[w0],
        verbose=verbose,
        normalizedUncer=normalizedUncer,
    )

    correction = np.ones(vis.shape)
    correction[w0] = np.exp(fit["model"] * corr_const)

    if display:
        plt.figure()
        plt.errorbar(
            (X[0] ** 2 + X[1] ** 2) ** 0.5,
            normvar[w0],
            yerr=err_avar[w0],
            ls="None",
            ecolor="lightgray",
            marker=".",
            label="data",
        )
        plt.plot((X[0] ** 2 + X[1] ** 2) ** 0.5, fit["model"], "rs")
        plt.show(block=False)

    return correction


def average_calib_files(list_nrm, sig_thres=2, display=False):
    """Average NRM data extracted from multiple calibrator files. Additionaly,
    perform sigma-clipping to reject suspicious dataset.

    Parameters:
    -----------
    `list_nrm` : {list}
        List of classes containing extracted NRM data (see bispect.py) of multiple calibrator files,\n
    `sig_thres` : {float}
        Threshold of the sigma clipping (default: 2-sigma around the median is used),\n
    """
    nfiles = len(list_nrm)
    l_pa = np.zeros(nfiles)
    cp_vs_file, e_cp_vs_file = [], []
    vis2_vs_file, e_vis2_vs_file = [], []

    # Fill array containing each vis2 and cp across files.
    for n in range(nfiles):
        nrm = list_nrm[n]
        hdu = fits.open(nrm.infos.filename)
        hdr = hdu[0].header
        try:
            # todo: Check parallactic angle param of a real NIRISS header.
            l_pa[n] = hdr["PARANG"]
        except KeyError:
            l_pa[n] = 0

        cp = nrm.cp
        e_cp = nrm.e_cp
        vis2 = nrm.vis2
        e_vis2 = nrm.e_vis2

        cp_vs_file.append(cp)
        e_cp_vs_file.append(e_cp)
        vis2_vs_file.append(vis2)
        e_vis2_vs_file.append(e_vis2)

    bl = list_nrm[0].bl

    cp_vs_file = np.array(cp_vs_file)
    e_cp_vs_file = np.array(e_cp_vs_file)
    vis2_vs_file = np.array(vis2_vs_file)
    e_vis2_vs_file = np.array(e_vis2_vs_file)

    zero_uncer = e_vis2_vs_file == 0
    e_vis2_vs_file[zero_uncer] = np.max(e_vis2_vs_file)

    cmn_vis2, std_vis2 = wtmn(vis2_vs_file, e_vis2_vs_file)
    cmn_cp, std_cp = wtmn(cp_vs_file, e_cp_vs_file)

    # Apply sigma clipping on the averages
    cmn_vis2_clip, std_vis2_clip = _apply_sig_clip(
        vis2_vs_file, e_vis2_vs_file, sig_thres=sig_thres, var="V2", display=display
    )
    cmn_cp_clip, std_cp_clip = _apply_sig_clip(
        cp_vs_file,
        e_cp_vs_file,
        sig_thres=sig_thres,
        ymax=10,
        var="CP",
        display=display,
    )

    res = {
        "f_v2_clip": np.array(cmn_vis2_clip),
        "f_v2": np.array(cmn_vis2),
        "std_vis2_clip": np.array(std_vis2_clip),
        "std_vis2": np.array(std_vis2),
        "f_cp_clip": np.array(cmn_cp_clip),
        "f_cp": np.array(cmn_cp),
        "std_cp_clip": np.array(std_cp_clip),
        "std_cp": np.array(std_cp),
        "bl": bl,
        "pa": l_pa,
    }

    return dict2class(res)


def calibrate(
    res_t,
    res_c,
    clip=False,
    sig_thres=2,
    apply_phscorr=False,
    apply_atmcorr=False,
    normalize_err_indep=False,
    display=False,
):
    """Calibrate v2 and cp from a science target and its calibrator.

    Parameters
    ----------
    `res_t` : {dict}
        Dictionnary containing extracted NRM data of science target (see bispect.py),\n
    `res_c` : {list or dict}
        Dictionnary or a list of dictionnary containing extracted NRM data of calibrator target,\n
    `clip` : {bool}
        If True, sigma clipping is performed over the calibrator files (if any) to reject bad
        observables due to seeing conditions, centering, etc.,\n
    `sig_thres` : {float}
        Threshold of the sigma clipping (default: 2-sigma around the median is used),\n
    `apply_phscorr` : {bool}, optional
        If True, apply a phasor correction due to piston between holes, by default False.\n
    `apply_atmcorr` : {bool}, optional
        If True, apply a atmospheric correction on V2 from seeing and wind shacking issues, by default False.\n
    `normalize_err_indep` : {bool}, optional
        If True, the CP uncertaintities are normalized by np.sqrt(n_holes/3.) to not over use
        the non-independant closure phases.\n
    `display`: {bool}
        If True, plot figures.


    Returns
    -------
    `cal`: {class}
        Class of calibrated data, keys are: `v2`, `e_v2` (squared visibities and errors),
        `cp`, `e_cp` (closure phase and errors), `visamp`, `e_visamp` (visibility
        ampliture and errors), `visphi`, `e_visphi` (visibility phase and errors), `u`, `v`
        (u-v coordinates), `wl` (wavelength), `raw_t` and `raw_c` (dictionnary of extracted raw
        NRM data, inputs of this function).
    """

    if type(res_c) is not list:
        res_c = [res_c]

    calib_tab = average_calib_files(res_c, sig_thres=sig_thres, display=display)

    if clip:
        cmn_v2_c, cmn_cp_c, std_v2_c, std_cp_c = (
            calib_tab.f_v2_clip,
            calib_tab.f_cp_clip,
            calib_tab.std_vis2_clip,
            calib_tab.std_cp_clip,
        )
    else:
        cmn_v2_c, cmn_cp_c, std_v2_c, std_cp_c = (
            calib_tab.f_v2,
            calib_tab.f_cp,
            calib_tab.std_vis2,
            calib_tab.std_cp,
        )

    v2_corr_t = 1
    if apply_atmcorr:
        v2_corr_t = _calc_correction_atm_vis2(res_t)

    if apply_phscorr:
        v2_corr_t *= res_t.matrix.phs_v2corr

    # Raw V2 target (corrected from atm correction and phasors.)
    v2_t = res_t.vis2 / v2_corr_t
    e_v2_t = res_t.e_vis2 / v2_corr_t
    #  e_v2_c = res_c[0].e_vis2/v2_corr_t

    # Raw CP target
    cp_t = res_t.cp
    e_cp_t = res_t.e_cp

    # Calibration by the weighted averages and taking into accound the std of the calibrators.
    # ---------------------------------------------
    vis2_calib = v2_t / cmn_v2_c
    cp_calib = cp_t - cmn_cp_c

    #
    n_holes = res_t.mask.n_holes
    if normalize_err_indep:
        err_scale = np.sqrt(n_holes / 3.0)
    else:
        err_scale = 1

    # Quadratic added error due to calibrator dispersion (the average is weightened (see wtmn from amical.tools)).
    weightened_error = True
    if weightened_error:
        e_vis2_calib = np.sqrt(
            e_v2_t ** 2 / cmn_v2_c ** 2 + std_v2_c ** 2 * v2_t ** 2 / cmn_v2_c ** 4
        )
    else:
        e_vis2_calib = np.sqrt(e_v2_t ** 2 + std_v2_c ** 2)

    e_cp_calib = np.sqrt(e_cp_t ** 2 + std_cp_c ** 2) * err_scale

    u1 = res_t.u[res_t.mask.bs2bl_ix[0, :]]
    v1 = res_t.v[res_t.mask.bs2bl_ix[0, :]]
    u2 = res_t.u[res_t.mask.bs2bl_ix[1, :]]
    v2 = res_t.v[res_t.mask.bs2bl_ix[1, :]]

    cal = {
        "vis2": vis2_calib,
        "e_vis2": e_vis2_calib,
        "cp": cp_calib,
        "e_cp": e_cp_calib,
        "u": res_t.u,
        "v": res_t.v,
        "wl": res_t.wl,
        "u1": u1,
        "v1": v1,
        "u2": u2,
        "v2": v2,
        "raw_t": res_t,
        "raw_c": res_c,
    }
    return dict2class(cal)



import copy
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import oifits
from matplotlib.patches import Rectangle
from matplotlib.transforms import offset_copy
from mpl_toolkits.mplot3d import Axes3D
from numpy import arctan, pi, sort, sqrt


def match_wavelength(template, oifitsobj):

    oifitsobj = copy.deepcopy(oifitsobj)
    oifitsobj.wavelength = template.wavelength

    for tvis, vis in zip(template.vis, oifitsobj.vis):
        idx = np.argsort(vis.wavelength.eff_wave)
        vis._visamp = np.interp(
            tvis.wavelength.eff_wave, vis.wavelength.eff_wave[idx], vis._visamp[idx]
        )
        vis._visamperr = np.interp(
            tvis.wavelength.eff_wave, vis.wavelength.eff_wave[idx], vis._visamperr[idx]
        )
        vis._visphi = np.interp(
            tvis.wavelength.eff_wave, vis.wavelength.eff_wave[idx], vis._visphi[idx]
        )
        vis._visphierr = np.interp(
            tvis.wavelength.eff_wave, vis.wavelength.eff_wave[idx], vis._visphierr[idx]
        )
        if vis._cflux != None:
            vis._cflux = np.interp(
                tvis.wavelength.eff_wave, vis.wavelength.eff_wave[idx], vis._cflux[idx]
            )
        if vis._cfluxerr != None:
            vis._visamperr = np.interp(
                tvis.wavelength.eff_wave,
                vis.wavelength.eff_wave[idx],
                vis._cfluxerr[idx],
            )
        vis.flag = np.zeros_like(vis._visamp, dtype=bool)
        vis.wavelength = tvis.wavelength

    return oifitsobj


def average_vector(vector, npix):

    newvect = np.zeros(vector.size / npix, dtype=vector.dtype)
    endpoint = vector.size - vector.size % npix
    for i in range(npix):
        newvect += vector[i:endpoint:npix] / float(npix)

    return newvect


def average_wavelength(oifitsobj, npix):

    source = copy.deepcopy(oifitsobj)

    for wavelength in source.wavelength.values():
        wavelength.eff_wave = average_vector(wavelength.eff_wave, npix)
        wavelength.eff_band = average_vector(wavelength.eff_band, npix)

    for vis in source.vis:
        vis._visamp = average_vector(vis._visamp, npix)
        vis._visamperr = average_vector(vis._visamperr, npix)
        vis._visphi = average_vector(vis._visphi, npix)
        vis._visphierr = average_vector(vis._visphierr, npix)
        if vis._cflux != None:
            vis._cflux = average_vector(vis._cflux, npix)
        if vis._cfluxerr != None:
            vis._cfluxerr = average_vector(vis._cfluxerr, npix)
        vis.flag = average_vector(vis.flag, npix)

    for vis2 in source.vis2:
        vis2._vis2data = average_vector(vis2._vis2data, npix)
        vis2._vis2err = average_vector(vis2._vis2err, npix)
        vis2.flag = average_vector(vis2.flag, npix)

    for t3 in source.t3:
        t3._t3amp = average_vector(t3._t3amp, npix)
        t3._t3amperr = average_vector(t3._t3amperr, npix)
        t3._t3phi = average_vector(t3._t3phi, npix)
        t3._t3phierr = average_vector(t3._t3phierr, npix)
        t3.flag = average_vector(t3.flag, npix)

    return source


def plot_fullvis(oifitsobj):

    baseline = np.empty(0)
    visamp = np.empty(0)
    visamperr = np.empty(0)

    for vis in oifitsobj.vis:
        baseline = np.append(
            baseline,
            (vis.ucoord ** 2 + vis.vcoord ** 2)
            / vis.wavelength.eff_wave[np.where(vis.flag == False)],
        )
        visamp = np.append(visamp, vis.visamp[np.where(vis.flag == False)])
        visamperr = np.append(visamperr, vis.visamperr[np.where(vis.flag == False)])

    # plt.errorbar(baseline, visamp, visamperr, marker='dot', linestyle='none', color='black')
    plt.plot(baseline, visamp, "k.")


def plot_full_visamp_vs_spatfreq(source, pmin=None, pmax=None, showerror=False):

    cmap = matplotlib.cm.rainbow

    fig = plt.figure()
    ax = fig.add_axes((0.125, 0.1, 0.9 - 0.125, 0.75))

    wave = source.vis[0].wavelength.eff_wave * 1e6

    posangles = np.empty(0)
    for vis in source.vis:
        posangle = np.arctan(vis.ucoord / vis.vcoord) * 180.0 / np.pi % 180.0
        posangles = np.append(posangles, posangle)

    if pmin == None:
        pmin = posangles.min()
    if pmax == None:
        pmax = posangles.max()

    for vis in source.vis:
        spatfreq = (
            np.sqrt(vis.ucoord ** 2 + vis.vcoord ** 2)
            / vis.wavelength.eff_wave
            / np.pi
            / 180.0
            / 3600.0
        )
        posangle = np.arctan(vis.ucoord / vis.vcoord) * 180.0 / np.pi
        posangle %= 180.0
        if showerror:
            ax.errorbar(
                spatfreq,
                vis.visamp,
                vis.visamperr,
                color=cmap((posangle - pmin) / (pmax - pmin)),
            )
        else:
            ax.plot(spatfreq, vis.visamp, color=cmap((posangle - pmin) / (pmax - pmin)))

    ax.set_xlabel("Spatial frequency (fringe cycles/arcsec)")
    ax.set_ylabel("Visibility amplitude")
    ax.set_xlim(0, None)
    ax.set_ylim(0, None)

    position = ax.get_position().get_points()
    cbarax = fig.add_axes(
        (position[0, 0], 0.9, position[1, 0] - position[0, 0], 0.05), frameon=False
    )
    cbarax.imshow(
        [np.linspace(0, 1, 100)], cmap=cmap, aspect="auto", extent=(pmin, pmax, 0, 1)
    )

    plt.setp(cbarax.get_xticklabels(), visible=True)
    plt.setp(cbarax.get_yticklabels(), visible=False)
    plt.setp(cbarax.get_xticklines(), visible=True)
    plt.setp(cbarax.get_yticklines(), visible=False)
    cbarax.set_title("Position angle (deg)")

    fig.show()
    return posangles


def get_gaussian_width_from_visibility(baselines, visibilities):

    sigma = np.sqrt(-np.log(visibilities) / 2.0) / np.pi / baselines
    fwhm = np.sqrt(8.0 * np.log(2.0)) * sigma
    fwhm *= 180.0 / np.pi * 3600.0

    return fwhm


def plot_visamp_vs_spatfreq(
    source, waveidx=None, pmin=None, pmax=None, fig=None, marker="o"
):

    cmap = matplotlib.cm.rainbow

    if waveidx == None:
        waveidx = np.arange(120, 125)

    if np.isscalar(waveidx):
        waveidx = np.array([waveidx])

    wave = source.vis[0].wavelength.eff_wave * 1e6

    posangles = np.empty(0)
    for vis in source.vis:
        posangle = np.arctan(vis.ucoord / vis.vcoord) * 180.0 / np.pi % 180.0
        posangles = np.append(posangles, posangle)

    if fig == None:
        fig = plt.figure()
        ax = fig.add_axes((0.125, 0.1, 0.9 - 0.125, 0.75))
        if pmin == None:
            pmin = posangles.min()
        if pmax == None:
            pmax = posangles.max()
        position = ax.get_position().get_points()
        cbarax = fig.add_axes(
            (position[0, 0], 0.9, position[1, 0] - position[0, 0], 0.05), frameon=False
        )
        # These can be saved into the fig object for recall later...
        fig.pmin = pmin
        fig.pmax = pmax
    else:
        ax = fig.axes[0]
        pmin = fig.pmin
        pmax = fig.pmax
        cbarax = fig.axes[1]

    for vis in source.vis:
        spatfreq = (
            np.sqrt(vis.ucoord ** 2 + vis.vcoord ** 2)
            / vis.wavelength.eff_wave[waveidx]
            / np.pi
            / 180.0
            / 3600.0
        ).mean()
        visamp = vis.visamp[waveidx].mean()
        visamperr = np.sqrt((vis.visamperr[waveidx] ** 2).sum() / np.size(waveidx))
        posangle = np.arctan(vis.ucoord / vis.vcoord) * 180.0 / np.pi
        if posangle < 0:
            posangle += 180.0
        ax.errorbar(
            [spatfreq],
            [visamp],
            [visamperr],
            color=cmap((posangle - pmin) / (pmax - pmin)),
            marker=marker,
        )

    ax.set_xlabel("Spatial frequency (fringe cycles/arcsec)")
    ax.set_ylabel("Visibility amplitude")
    ax.text(
        0.98,
        0.98,
        r"$\lambda=%.1f \mu$m" % (wave[waveidx].mean()),
        ha="right",
        va="top",
        transform=ax.transAxes,
    )
    ax.set_xlim(0, None)
    ax.set_ylim(0, None)

    position = ax.get_position().get_points()
    cbarax = fig.add_axes(
        (position[0, 0], 0.9, position[1, 0] - position[0, 0], 0.05), frameon=False
    )
    cbarax.imshow(
        [np.linspace(0, 1, 100)], cmap=cmap, aspect="auto", extent=(pmin, pmax, 0, 1)
    )

    plt.setp(cbarax.get_xticklabels(), visible=True)
    plt.setp(cbarax.get_yticklabels(), visible=False)
    plt.setp(cbarax.get_xticklines(), visible=True)
    plt.setp(cbarax.get_yticklines(), visible=False)
    cbarax.set_title("Position angle (deg)")

    return fig


def plot_gaussian_widths(source, waveidx=100, legend=False):

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    for vis in source.vis:
        label = vis.station[0].sta_name + vis.station[1].sta_name
        baseline = np.sqrt(vis.ucoord ** 2 + vis.vcoord ** 2)
        fwhm = get_gaussian_width_from_visibility(
            baseline / vis.wavelength.eff_wave[waveidx], vis.visamp[waveidx]
        )
        fwhm_upper = get_gaussian_width_from_visibility(
            baseline / vis.wavelength.eff_wave[waveidx],
            vis.visamp[waveidx] - vis.visamperr[waveidx],
        )
        fwhm_lower = get_gaussian_width_from_visibility(
            baseline / vis.wavelength.eff_wave[waveidx],
            vis.visamp[waveidx] + vis.visamperr[waveidx],
        )
        line = ax.plot([baseline], [fwhm], ".", label=label)
        ax.errorbar([baseline], [fwhm], [fwhm_upper - fwhm], color=line[0].get_color())
    if legend:
        plt.legend()
    ax.set_xlabel("Baseline (m)")
    ax.set_ylabel("FWHM (arcsec)")
    ax.set_title("%g $\\mu$m" % (source.vis[0].wavelength.eff_wave[waveidx] * 1e6))
    fig.show()
    return fig


def get_ranges(data, value):

    ranges = []
    inrange = False
    i = 0

    while i < len(data):
        if data[i] == value and inrange == False:
            start = i
            inrange = True
        elif data[i] != value and inrange == True:
            ranges.append((start, i - 1))
            inrange = False
        elif i == len(data) - 1 and inrange == True:
            ranges.append((start, i))
        i += 1

    return ranges


def plot_visamp_map(oifitsobj, width=0.1, height=0.1, colorcoding=None):

    fig = plt.figure(figsize=(10, 8), facecolor="white")

    for vis in oifitsobj.vis:
        B = sqrt(vis.ucoord ** 2 + vis.vcoord ** 2)
        pa = (arctan(vis.ucoord / vis.vcoord) * 180.0 / pi) % 180.0
        print(B, pa)
        ax = fig.add_axes(
            (B / 80.0 - width / 2.0, pa / 180.0 - height / 2.0, width, height),
            axis_bgcolor=(0, 0, 0, 0),
        )
        if colorcoding:
            for key in sort(colorcoding.keys()):
                if B < key:
                    color = colorcoding[key]
                    break
        else:
            color = "k"

        ax.errorbar(
            1e6 * vis.wavelength.eff_wave,
            vis.visamp,
            vis.visamperr,
            fmt="k-",
            color=color,
        )
        ax.set_xlim(8, 13)
        ax.set_ylim(0, 0.8)
        # ax.xaxis.set_visible(False)
        # ax.yaxis.set_visible(False)
        ax.xaxis.set_ticklabels(
            (),
        )
        ax.yaxis.set_ticklabels(
            (),
        )

    return fig


def plot_phases(oidata, uvplot=False, legend=False):

    xmin = None
    xmax = None
    ymin = 0.0
    ymax = 0.0
    baselinemax = None
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    if uvplot:
        fig2 = plt.figure(figsize=(5, 5))
        ax2 = fig2.add_subplot(111)

    for vis in oidata.vis:
        ranges = get_ranges(vis.flag, True)
        for bounds in ranges:
            w1 = vis.wavelength.eff_wave[bounds[0]] * 1e6
            w2 = vis.wavelength.eff_wave[bounds[1]] * 1e6
            rect = matplotlib.patches.Rectangle(
                (w1, -180), w2 - w1, 360, color="lightblue"
            )
            ax1.add_patch(rect)

    names = []

    colors = (
        "black",
        "red",
        "green",
        "blue",
        "purple",
        "teal",
        "orange",
        "cyan",
        "darkred",
        "darkgreen",
        "darkblue",
        "darkorange",
        "grey",
        "darkgrey",
        "lightgrey",
    )
    colorid = 0

    for vis in oidata.vis:
        u = vis.ucoord
        v = vis.vcoord
        if vis.station[0] and vis.station[1]:
            label = vis.station[0].sta_name + vis.station[1].sta_name
        else:
            label = "unnamed"
        color = colors[colorid]
        colorid = np.mod(colorid + 1, len(colors))
        line = ax1.errorbar(
            1e6 * vis.wavelength.eff_wave,
            vis.visphi,
            vis.visphierr,
            label=label,
            color=color,
        )

        ymin = np.amin(np.append(vis.visphi[np.where(vis.flag == False)], ymin))
        ymax = np.amax(np.append(vis.visphi[np.where(vis.flag == False)], ymax))
        xmin = np.amin(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmax)
        )
        xmax = np.amax(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmin)
        )
        if uvplot:
            ax2.plot([-u, u], [-v, v], ".", label=label, color=color)
        names.append(vis.target.target)
        baselinemax = np.amax([np.sqrt(u ** 2 + v ** 2), baselinemax])
        print(
            "%10s: %s, %5.2f m, %s"
            % (line[0].get_color(), label, np.sqrt(u ** 2 + v ** 2), vis.timeobs)
        )

    ax1.set_xlim(xmin, xmax)
    ax1.set_ylim(ymin, ymax)
    if uvplot:
        ax2.set_xlim(-1.1 * baselinemax, 1.1 * baselinemax)
        ax2.set_ylim(-1.1 * baselinemax, 1.1 * baselinemax)

    names = list(np.unique(names))
    title = names.pop()
    for name in names:
        title += ", %s" % (name)

    ax1.set_title(title)
    ax1.set_xlabel(r"Wavelength ($\mu$m)")
    ax1.set_ylabel("Differential phase")
    if uvplot:
        ax2.set_title(title)
        ax2.set_xlabel("u (m)")
        ax2.set_ylabel("v (m)")

    if legend and uvplot:
        ax2.legend(prop={"size": 10}, numpoints=1)


def plot_visibilities(oidata, uvplot=False, legend=False, ploterror=False):

    xmin = None
    xmax = None
    ymax = None
    baselinemax = None
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    if uvplot:
        fig2 = plt.figure(figsize=(6, 6))
        ax2 = fig2.add_subplot(111)

    for vis in oidata.vis:
        ranges = get_ranges(vis.flag, True)
        for bounds in ranges:
            w1 = vis.wavelength.eff_wave[bounds[0]] * 1e6
            w2 = vis.wavelength.eff_wave[bounds[1]] * 1e6
            rect = matplotlib.patches.Rectangle(
                (w1, 0), w2 - w1, 999, color="lightblue"
            )
            ax1.add_patch(rect)

    names = []

    colors = (
        "black",
        "red",
        "green",
        "blue",
        "purple",
        "teal",
        "orange",
        "cyan",
        "darkred",
        "darkgreen",
        "darkblue",
        "darkorange",
        "grey",
        "darkgrey",
        "lightgrey",
        "pink",
    )
    colorid = 0
    output = ""

    for vis in oidata.vis:
        u = vis.ucoord
        v = vis.vcoord
        if vis.station[0] and vis.station[1]:
            label = vis.station[0].sta_name + vis.station[1].sta_name
        else:
            label = "unnamed"
        color = colors[colorid]
        colorid = np.mod(colorid + 1, len(colors))
        try:
            if ploterror:
                line = ax1.errorbar(
                    1e6 * vis.wavelength.eff_wave,
                    vis.visamp,
                    vis.visamperr,
                    label=label,
                    color=color,
                )
            else:
                line = ax1.plot(
                    1e6 * vis.wavelength.eff_wave, vis.visamp, label=label, color=color
                )
        except:
            if ploterror:
                line = ax1.errorbar(
                    1e6 * vis.wavelength.eff_wave,
                    vis.vis2data,
                    vis.vis2err,
                    label=label,
                    color=color,
                )
            else:
                line = ax1.plot(
                    1e6 * vis.wavelength.eff_wave,
                    vis.vis2data,
                    label=label,
                    color=color,
                )
        ymax = np.amax(np.append(vis.visamp.max(), ymax))
        xmin = np.amin(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmax)
        )
        xmax = np.amax(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmin)
        )
        if uvplot:
            ax2.plot([-u, u], [-v, v], ".", label=label, color=color)
        names.append(vis.target.target)
        baselinemax = np.amax([np.sqrt(u ** 2 + v ** 2), baselinemax])
        output += "{:>10}: {}, {:5.2f} m, {}\n".format(
            line[0].get_color(),
            label,
            np.sqrt(u ** 2 + v ** 2),
            vis.timeobs,
        )

    ax1.set_xlim(xmin, xmax)
    ax1.set_ylim(0, ymax)
    if uvplot:
        ax2.set_xlim(-1.1 * baselinemax, 1.1 * baselinemax)
        ax2.set_ylim(-1.1 * baselinemax, 1.1 * baselinemax)

    names = list(np.unique(names))
    title = names.pop()
    for name in names:
        title += ", %s" % (name)

    ax1.set_title(title)
    ax1.set_xlabel(r"Wavelength ($\mu$m)")
    ax1.set_ylabel("Visibility")
    if uvplot:
        ax2.set_title(title)
        ax2.set_xlabel("u (m)")
        ax2.set_ylabel("v (m)")

    if legend and uvplot:
        ax2.legend(prop={"size": 10}, numpoints=1)
    print(output)


def plot_gaussian_widths_vs_wavelength(
    oidata, uvplot=False, legend=False, ploterror=False
):

    xmin = None
    xmax = None
    ymax = None
    baselinemax = None
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    if uvplot:
        fig2 = plt.figure(figsize=(6, 6))
        ax2 = fig2.add_subplot(111)

    for vis in oidata.vis:
        ranges = get_ranges(vis.flag, True)
        for bounds in ranges:
            w1 = vis.wavelength.eff_wave[bounds[0]] * 1e6
            w2 = vis.wavelength.eff_wave[bounds[1]] * 1e6
            rect = matplotlib.patches.Rectangle(
                (w1, 0), w2 - w1, 999, color="lightblue"
            )
            ax1.add_patch(rect)

    names = []

    colors = (
        "black",
        "red",
        "green",
        "blue",
        "purple",
        "teal",
        "orange",
        "cyan",
        "darkred",
        "darkgreen",
        "darkblue",
        "darkorange",
        "grey",
        "darkgrey",
        "lightgrey",
        "pink",
    )
    colorid = 0
    output = ""

    for vis in oidata.vis:
        u = vis.ucoord
        v = vis.vcoord
        if vis.station[0] and vis.station[1]:
            label = vis.station[0].sta_name + vis.station[1].sta_name
        else:
            label = "unnamed"
        color = colors[colorid]
        colorid = np.mod(colorid + 1, len(colors))
        widths = 1e3 * get_gaussian_width_from_visibility(
            sqrt(vis.ucoord ** 2 + vis.vcoord ** 2) / vis.wavelength.eff_wave,
            vis.visamp,
        )
        line = ax1.plot(1e6 * vis.wavelength.eff_wave, widths, label=label, color=color)
        ymax = np.amax(np.append(widths.max(), ymax))
        xmin = np.amin(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmax)
        )
        xmax = np.amax(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmin)
        )
        if uvplot:
            ax2.plot([-u, u], [-v, v], ".", label=label, color=color)
        names.append(vis.target.target)
        baselinemax = np.amax([np.sqrt(u ** 2 + v ** 2), baselinemax])
        output += "{:>10}: {}, {:5.2f} m, {}\n".format(
            line[0].get_color(),
            label,
            np.sqrt(u ** 2 + v ** 2),
            vis.timeobs,
        )

    ax1.set_xlim(xmin, xmax)
    ax1.set_ylim(0, ymax)
    if uvplot:
        ax2.set_xlim(-1.1 * baselinemax, 1.1 * baselinemax)
        ax2.set_ylim(-1.1 * baselinemax, 1.1 * baselinemax)

    names = list(np.unique(names))
    title = names.pop()
    for name in names:
        title += ", %s" % (name)

    ax1.set_title(title)
    ax1.set_xlabel(r"Wavelength ($\mu$m)")
    ax1.set_ylabel("Gaussian width (mas)")
    if uvplot:
        ax2.set_title(title)
        ax2.set_xlabel("u (m)")
        ax2.set_ylabel("v (m)")

    if legend and uvplot:
        ax2.legend(prop={"size": 10}, numpoints=1)
    print(output)


def plot_cflux(oidata, uvplot=False, legend=False, ploterror=False):

    xmin = None
    xmax = None
    ymax = None
    baselinemax = None
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    if uvplot:
        fig2 = plt.figure(figsize=(6, 6))
        ax2 = fig2.add_subplot(111)

    for vis in oidata.vis:
        ranges = get_ranges(vis.flag, True)
        for bounds in ranges:
            w1 = vis.wavelength.eff_wave[bounds[0]] * 1e6
            w2 = vis.wavelength.eff_wave[bounds[1]] * 1e6
            rect = matplotlib.patches.Rectangle(
                (w1, 0), w2 - w1, 999, color="lightblue"
            )
            ax1.add_patch(rect)

    names = []

    colors = (
        "black",
        "red",
        "green",
        "blue",
        "purple",
        "teal",
        "orange",
        "cyan",
        "darkred",
        "darkgreen",
        "darkblue",
        "darkorange",
        "grey",
        "darkgrey",
        "lightgrey",
        "pink",
    )
    colorid = 0
    output = ""

    for vis in oidata.vis:
        u = vis.ucoord
        v = vis.vcoord
        if vis.station[0] and vis.station[1]:
            label = vis.station[0].sta_name + vis.station[1].sta_name
        else:
            label = "unnamed"
        color = colors[colorid]
        colorid = np.mod(colorid + 1, len(colors))
        if ploterror:
            line = ax1.errorbar(
                1e6 * vis.wavelength.eff_wave,
                vis.cflux,
                vis.cfluxerr,
                label=label,
                color=color,
            )
        else:
            line = ax1.plot(
                1e6 * vis.wavelength.eff_wave, vis.cflux, label=label, color=color
            )
        ymax = np.amax(np.append(vis.cflux.max(), ymax))
        xmin = np.amin(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmax)
        )
        xmax = np.amax(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmin)
        )
        if uvplot:
            ax2.plot([-u, u], [-v, v], ".", label=label, color=color)
        names.append(vis.target.target)
        baselinemax = np.amax([np.sqrt(u ** 2 + v ** 2), baselinemax])
        output += "{:>10}: {}, {:5.2f} m, {}\n".format(
            line[0].get_color(),
            label,
            np.sqrt(u ** 2 + v ** 2),
            vis.timeobs,
        )

    ax1.set_xlim(xmin, xmax)
    ax1.set_ylim(0, ymax)
    if uvplot:
        ax2.set_xlim(-1.1 * baselinemax, 1.1 * baselinemax)
        ax2.set_ylim(-1.1 * baselinemax, 1.1 * baselinemax)

    names = list(np.unique(names))
    title = names.pop()
    for name in names:
        title += ", %s" % (name)

    ax1.set_title(title)
    ax1.set_xlabel(r"Wavelength ($\mu$m)")
    ax1.set_ylabel("Correlated flux (Jy)")
    if uvplot:
        ax2.set_title(title)
        ax2.set_xlabel("u (m)")
        ax2.set_ylabel("v (m)")

    if legend and uvplot:
        ax2.legend(prop={"size": 10}, numpoints=1)
    print(output)


def plot_vis2(oidata, uvplot=False, legend=False, ploterror=False):

    xmin = None
    xmax = None
    ymax = None
    baselinemax = None
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    if uvplot:
        fig2 = plt.figure(figsize=(6, 6))
        ax2 = fig2.add_subplot(111)

    for vis in oidata.vis2:
        ranges = get_ranges(vis.flag, True)
        for bounds in ranges:
            w1 = vis.wavelength.eff_wave[bounds[0]] * 1e6
            w2 = vis.wavelength.eff_wave[bounds[1]] * 1e6
            rect = matplotlib.patches.Rectangle(
                (w1, 0), w2 - w1, 999, color="lightblue"
            )
            ax1.add_patch(rect)

    names = []

    colors = (
        "black",
        "red",
        "green",
        "blue",
        "purple",
        "teal",
        "orange",
        "cyan",
        "darkred",
        "darkgreen",
        "darkblue",
        "darkorange",
        "grey",
        "darkgrey",
        "lightgrey",
        "pink",
    )
    colorid = 0

    for vis in oidata.vis2:
        u = vis.ucoord
        v = vis.vcoord
        if vis.station[0] and vis.station[1]:
            label = vis.station[0].sta_name + vis.station[1].sta_name
        else:
            label = "unnamed"
        color = colors[colorid]
        colorid = np.mod(colorid + 1, len(colors))
        if ploterror:
            line = ax1.errorbar(
                1e6 * vis.wavelength.eff_wave,
                vis.vis2data,
                vis.vis2err,
                label=label,
                color=color,
            )
        else:
            print(1e6 * vis.wavelength.eff_wave, vis.vis2data)
            line = ax1.plot(
                1e6 * vis.wavelength.eff_wave, vis.vis2data, label=label, color=color
            )
        ymax = np.amax(np.append(vis.vis2data[np.where(vis.flag == False)], ymax))
        xmin = np.amin(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmax)
        )
        xmax = np.amax(
            np.append(1e6 * vis.wavelength.eff_wave[np.where(vis.flag == False)], xmin)
        )
        if uvplot:
            ax2.plot([-u, u], [-v, v], ".", label=label, color=color)
        names.append(vis.target.target)
        baselinemax = np.amax([np.sqrt(u ** 2 + v ** 2), baselinemax])
        print(
            "%10s: %s, %5.2f m, %s"
            % (line[0].get_color(), label, np.sqrt(u ** 2 + v ** 2), vis.timeobs)
        )

    ax1.set_xlim(0.9 * xmin, 1.1 * xmax)
    ax1.set_ylim(0, ymax)
    if uvplot:
        ax2.set_xlim(-1.1 * baselinemax, 1.1 * baselinemax)
        ax2.set_ylim(-1.1 * baselinemax, 1.1 * baselinemax)

    names = list(np.unique(names))
    title = names.pop()
    for name in names:
        title += ", %s" % (name)

    ax1.set_title(title)
    ax1.set_xlabel(r"Wavelength ($\mu$m)")
    ax1.set_ylabel("Visibility")
    if uvplot:
        ax2.set_title(title)
        ax2.set_xlabel("u (m)")
        ax2.set_ylabel("v (m)")

    if legend and uvplot:
        ax2.legend(prop={"size": 10}, numpoints=1)


def print_vis(vis):

    print(
        "# %s %s %s"
        % (
            vis.target.target,
            vis.station[0].sta_name + vis.station[1].sta_name,
            vis.timeobs,
        )
    )
    print(f"# u = {vis.ucoord:6.2f} v = {vis.vcoord:6.2f}")
    print("# Wavelength (um)       ")
    print("#        |V|")
    print("#                |V|_err")

    idx = list(np.where(vis.flag == False)[0])
    idx.reverse()

    for i in idx:
        print(
            "%6.3f %8.4f %8.4f"
            % (1e6 * vis.wavelength.eff_wave[i], vis.visamp[i], vis.visamperr[i])
        )


def plot_array(array, visibilities=None):

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111)

    transOffset = offset_copy(
        ax.transData, fig=ax.figure, x=0.0, y=-0.10, units="inches"
    )

    center = array.arrxyz

    # Calculate a "east" unit vector
    east = np.array([center[1] * center[2], -center[0] * center[2], 0])
    if center[2] > 0:
        east *= -1
    east /= np.sqrt(np.sum(np.square(east)))
    # Calculate a "north" unit vector
    north = np.cross(center, east)
    north /= np.sqrt(np.sum(np.square(north)))

    stations = array.station
    xlist = []
    ylist = []

    for station in stations:
        if "U" in station.sta_name:
            color = "green"
        else:
            color = "blue"
        x = np.inner(station.staxyz, east)
        y = np.inner(station.staxyz, north)
        xlist.append(x)
        ylist.append(y)
        ax.plot([x], [y], "o", color=color, markersize=1 + station.diameter)
        plt.text(
            x,
            y,
            station.sta_name,
            transform=transOffset,
            horizontalalignment="center",
            verticalalignment="top",
            family="serif",
        )

    if visibilities:
        for vis in visibilities:
            x = np.array(
                [
                    np.inner(vis.station[0].staxyz, east),
                    np.inner(vis.station[1].staxyz, east),
                ]
            )
            y = np.array(
                [
                    np.inner(vis.station[0].staxyz, north),
                    np.inner(vis.station[1].staxyz, north),
                ]
            )
            ax.plot(
                x,
                y,
                linestyle="-",
                marker="|",
                markersize=20,
                label=vis.station[0].sta_name + vis.station[1].sta_name,
            )

    ax.plot([0], [0], "r+")
    plt.text(
        0,
        0,
        "$\\vec{O}$",
        transform=transOffset,
        horizontalalignment="center",
        verticalalignment="top",
    )

    ax.annotate(
        "N",
        xy=(0.05, 0.25),
        xytext=(0.05, 0.05),
        xycoords="axes fraction",
        textcoords="axes fraction",
        arrowprops={"width": 2},
        horizontalalignment="center",
        verticalalignment="bottom",
        family="serif",
        size="20",
    )

    minx = np.min(xlist)
    miny = np.min(ylist)
    maxx = np.max(xlist)
    maxy = np.max(ylist)
    centerx = (maxx - minx) / 2.0 + minx
    centery = (maxy - miny) / 2.0 + miny

    width = 1.1 * np.max([maxx - minx, maxy - miny])

    ax.set_xlim(centerx - width / 2.0, centerx + width / 2.0)
    ax.set_ylim(centery - width / 2.0, centery + width / 2.0)
    ax.relim()

    ax.set_xlabel("Relative position (m)")
    ax.set_ylabel("Relative position (m)")


def plot_uv_size(oidata, maxvis=0.3, waveidx=80):

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111)

    size = lambda visamp: visamp * 15 / maxvis + 1

    for vis in oidata.vis:
        ax.plot([vis.ucoord], [vis.vcoord], marker="o", ms=size(vis.visamp[waveidx]))
        ax.plot([-vis.ucoord], [-vis.vcoord], marker="o", ms=size(vis.visamp[waveidx]))

    for visamp in np.linspace(0.0, maxvis, 10):
        ax.plot([-1000], [-1000], "ko", ms=size(visamp), label="%.2g" % visamp)

    for radius in [10, 20, 30, 40, 50, 60]:
        x = np.linspace(-radius, radius, 100.0)
        y = np.sqrt(radius ** 2 - x ** 2)
        ax.plot(x, y, "k:")
        ax.plot(x, -y, "k:")

    ax.legend()

    ax.set_xlim(70, -70)
    ax.set_ylim(-70, 70)
    ax.set_xlabel("u (m)")
    ax.set_ylabel("v (m)")
    ax.set_title(
        r"$\lambda = %g$ $\mu$m" % (1e6 * oidata.vis[0].wavelength.eff_wave[waveidx])
    )


def station_distance(array, sta1name, sta2name):

    sta1 = array.get_station_by_name(sta1name)
    sta2 = array.get_station_by_name(sta2name)

    return sqrt(((sta1.staxyz - sta2.staxyz) ** 2).sum())



"""
@author: Anthony Soulain (University of Sydney)

-------------------------------------------------------------------------
AMICAL: Aperture Masking Interferometry Calibration and Analysis Library
-------------------------------------------------------------------------

Fitting tools (developped by A. Merand).

--------------------------------------------------------------------
"""

import time

import numpy as np
import scipy.optimize

"""
IDEA: fit Y = F(X,A) where A is a dictionnary describing the
parameters of the function.

note that the items in the dictionnary should all be scalar!

author: amerand@eso.org

Tue 29 Jan 2013 17:03:21 CLST: working on adding correlations -> NOT WORKING!!!
Thu 28 Feb 2013 12:34:31 CLST: correcting leading to x2 for chi2 display
Mon  8 Apr 2013 10:51:03 BRT: alternate algorithms

http://www.rhinocerus.net/forum/lang-idl-pvwave/355826-generalized-least-squares.html
"""

verboseTime = time.time()


def leastsqFit(
    func,
    x,
    params,
    y,
    err=None,
    fitOnly=None,
    verbose=False,
    doNotFit=None,
    epsfcn=1e-7,
    ftol=1e-5,
    fullOutput=True,
    normalizedUncer=True,
    follow=None,
):
    """
    - params is a Dict containing the first guess.

    - fits 'y +- err = func(x,params)'. errors are optionnal. in case err is a
      ndarray of 2 dimensions, it is treated as the covariance of the
      errors.

      np.array([[err1**2, 0, .., 0],
                [0, err2**2, 0, .., 0],
                [0, .., 0, errN**2]]) is the equivalent of 1D errors

    - follow=[...] list of parameters to "follow" in the fit, i.e. to print in
      verbose mode

    - fitOnly is a LIST of keywords to fit. By default, it fits all
      parameters in 'params'. Alternatively, one can give a list of
      parameters not to be fitted, as 'doNotFit='

    - doNotFit has a similar purpose: for example if params={'a0':,
      'a1': 'b1':, 'b2':}, doNotFit=['a'] will result in fitting only
      'b1' and 'b2'. WARNING: if you name parameter 'A' and another one 'AA',
      you cannot use doNotFit to exclude only 'A' since 'AA' will be excluded as
      well...

    - normalizedUncer=True: the uncertainties are independent of the Chi2, in
      other words the uncertainties are scaled to the Chi2. If set to False, it
      will trust the values of the error bars: it means that if you grossely
      underestimate the data's error bars, the uncertainties of the parameters
      will also be underestimated (and vice versa).

    returns dictionary with:
    'best': bestparam,
    'uncer': uncertainties,
    'chi2': chi2_reduced,
    'model': func(x, bestparam)
    'cov': covariance matrix (normalized if normalizedUncer)
    'fitOnly': names of the columns of 'cov'
    """

    if doNotFit is None:
        doNotFit = []

    # -- fit all parameters by default
    if fitOnly is None:
        if len(doNotFit) > 0:
            fitOnly = [x for x in list(params.keys()) if x not in doNotFit]
        else:
            fitOnly = list(params.keys())
        fitOnly.sort()  # makes some display nicer

    # -- build fitted parameters vector:
    pfit = [params[k] for k in fitOnly]

    # -- built fixed parameters dict:
    pfix = {}
    for k in list(params.keys()):
        if k not in fitOnly:
            pfix[k] = params[k]
    if verbose:
        print("[dpfit] %d FITTED parameters:" % (len(fitOnly)), fitOnly)
    # -- actual fit
    plsq, cov, info, mesg, ier = scipy.optimize.leastsq(
        _fitFunc,
        pfit,
        args=(
            fitOnly,
            x,
            y,
            err,
            func,
            pfix,
            verbose,
            follow,
        ),
        full_output=True,
        epsfcn=epsfcn,
        ftol=ftol,
    )
    if isinstance(err, np.ndarray) and len(err.shape) == 2:
        print(cov)

    # -- best fit -> agregate to pfix
    for i, k in enumerate(fitOnly):
        pfix[k] = plsq[i]

    # -- reduced chi2
    model = func(x, pfix)
    tmp = _fitFunc(plsq, fitOnly, x, y, err, func, pfix)

    try:
        chi2 = (np.array(tmp) ** 2).sum()
    except Exception:
        chi2 = 0.0
        for x in tmp:
            chi2 += np.sum(x ** 2)
    reducedChi2 = chi2 / float(
        np.sum([1 if np.isscalar(i) else len(i) for i in tmp]) - len(pfit) + 1
    )
    # print(chi2, reducedChi2, float(np.sum([1 if np.isscalar(i) else
    #                                 len(i) for i in tmp])-len(pfit)+1))
    if not np.isscalar(reducedChi2):
        reducedChi2 = np.mean(reducedChi2)

    # -- uncertainties:
    uncer = {}
    for k in list(pfix.keys()):
        if k not in fitOnly:
            uncer[k] = 0  # not fitted, uncertatinties to 0
        else:
            i = fitOnly.index(k)
            if cov is None:
                uncer[k] = -1
            else:
                uncer[k] = np.sqrt(np.abs(np.diag(cov)[i]))
                if normalizedUncer:
                    uncer[k] *= np.sqrt(reducedChi2)

    if verbose:
        print("-" * 30)
        print("REDUCED CHI2=", reducedChi2)
        print("-" * 30)
        if normalizedUncer:
            print("(uncertainty normalized to data dispersion)")
        else:
            print("(uncertainty assuming error bars are correct)")
        tmp = list(pfix.keys())
        tmp.sort()
        maxLength = np.max(np.array([len(k) for k in tmp]))
        format_ = "'%s':"
        # -- write each parameter and its best fit, as well as error
        # -- writes directly a dictionnary
        print("")  # leave some space to the eye
        for ik, k in enumerate(tmp):
            padding = " " * (maxLength - len(k))
            formatS = format_ + padding
            if ik == 0:
                formatS = "{" + formatS
            if uncer[k] > 0:
                ndigit = -int(np.log10(uncer[k])) + 3
                print(formatS % k, round(pfix[k], ndigit), ",", end=" ")
                print("# +/-", round(uncer[k], ndigit))
            elif uncer[k] == 0:
                if isinstance(pfix[k], str):
                    print(formatS % k, "'" + pfix[k] + "'", ",")
                else:
                    print(formatS % k, pfix[k], ",")
            else:
                print(formatS % k, pfix[k], ",", end=" ")
                print("# +/-", uncer[k])
        print("}")  # end of the dictionnary
        try:
            if verbose > 1:
                print("-" * 3, "correlations:", "-" * 15)
                N = np.max([len(k) for k in fitOnly])
                N = min(N, 20)
                N = max(N, 5)
                sf = "%" + str(N) + "s"
                print(" " * N, end=" ")
                for k2 in fitOnly:
                    print(sf % k2, end=" ")
                print("")
                sf = "%-" + str(N) + "s"
                for k1 in fitOnly:
                    i1 = fitOnly.index(k1)
                    print(sf % k1, end=" ")
                    for k2 in fitOnly:
                        i2 = fitOnly.index(k2)
                        if k1 != k2:
                            print(
                                ("%" + str(N) + ".2f")
                                % (cov[i1, i2] / np.sqrt(cov[i1, i1] * cov[i2, i2])),
                                end=" ",
                            )
                        else:
                            print(" " * (N - 4) + "-" * 4, end=" ")
                    print("")
                print("-" * 30)
        except Exception:
            pass
    # -- result:
    if fullOutput:
        if normalizedUncer:
            try:
                cov *= reducedChi2
            except Exception:
                pass
        try:
            cor = np.sqrt(np.diag(cov))
            cor = cor[:, None] * cor[None, :]
            cor = cov / cor
        except Exception:
            cor = None

        pfix = {
            "best": pfix,
            "uncer": uncer,
            "chi2": reducedChi2,
            "model": model,
            "cov": cov,
            "fitOnly": fitOnly,
            "info": info,
            "cor": cor,
        }
    return pfix


def bootstrap(
    func,
    x,
    params,
    y,
    err=None,
    fitOnly=None,
    verbose=False,
    doNotFit=None,
    epsfcn=1e-7,
    ftol=1e-5,
    fullOutput=True,
    normalizedUncer=True,
    follow=None,
    Nboot=None,
):
    """
    bootstraping, called like leastsqFit. returns a list of fits: the first one
    is the 'normal' one, the Nboot following one are with ramdomization of data. If
    Nboot is not given, it is set to 10*len(x).
    """
    if doNotFit is None:
        doNotFit = []

    if Nboot is None:
        Nboot = 10 * len(x)
    # first fit is the "normal" one
    fits = [
        leastsqFit(
            func,
            x,
            params,
            y,
            err=err,
            fitOnly=fitOnly,
            verbose=False,
            doNotFit=doNotFit,
            epsfcn=epsfcn,
            ftol=ftol,
            fullOutput=True,
            normalizedUncer=True,
        )
    ]
    for _ in range(Nboot):
        s = np.int_(len(x) * np.random.rand(len(x)))
        fits.append(
            leastsqFit(
                func,
                x[s],
                params,
                y[s],
                err=err,
                fitOnly=fitOnly,
                verbose=False,
                doNotFit=doNotFit,
                epsfcn=epsfcn,
                ftol=ftol,
                fullOutput=True,
                normalizedUncer=True,
            )
        )
    return fits


def _fitFunc(
    pfit, pfitKeys, x, y, err=None, func=None, pfix=None, verbose=False, follow=None
):
    """
    interface to leastsq from scipy:
    - x,y,err are the data to fit: f(x) = y +- err
    - pfit is a list of the paramters
    - pfitsKeys are the keys to build the dict
    pfit and pfix (optional) and combines the two
    in 'A', in order to call F(X,A)

    in case err is a ndarray of 2 dimensions, it is treated as the
    covariance of the errors.
    np.array([[err1**2, 0, .., 0],
             [ 0, err2**2, 0, .., 0],
             [0, .., 0, errN**2]]) is the equivalent of 1D errors

    """
    global verboseTime
    params = {}
    # -- build dic from parameters to fit and their values:
    for i, k in enumerate(pfitKeys):
        params[k] = pfit[i]
    # -- complete with the non fitted parameters:
    for k in pfix:
        params[k] = pfix[k]
    if err is None:
        err = np.ones(np.array(y).shape)

    # -- compute residuals

    if type(y) == np.ndarray and type(err) == np.ndarray:
        if len(err.shape) == 2:
            # -- using correlations
            tmp = func(x, params)
            # res = np.dot(np.dot(tmp-y, linalg.inv(err)), tmp-y)
            res = np.dot(np.dot(tmp - y, err), tmp - y)
            res = np.ones(len(y)) * np.sqrt(res / len(y))
        else:
            # -- assumes y and err are a numpy array
            y = np.array(y)
            res = ((func(x, params) - y) / err).flatten()
    else:
        # much slower: this time assumes y (and the result from func) is
        # a list of things, each convertible in np.array
        res = []
        tmp = func(x, params)

        for k in range(len(y)):
            df = (np.array(tmp[k]) - np.array(y[k])) / np.array(err[k])
            try:
                res.extend(list(df))
            except Exception:
                res.append(df)

    if verbose and time.time() > (verboseTime + 1):
        verboseTime = time.time()
        print(time.asctime(), end=" ")
        try:
            chi2 = (res ** 2).sum / (len(res) - len(pfit) + 1.0)
            print("CHI2: %6.4e" % chi2, end=" ")
        except Exception:
            # list of elements
            chi2 = 0
            N = 0
            res2 = []
            for r in res:
                if np.isscalar(r):
                    chi2 += r ** 2
                    N += 1
                    res2.append(r)
                else:
                    chi2 += np.sum(np.array(r) ** 2)
                    N += len(r)
                    res2.extend(list(r))

            res = res2
            print("CHI2: %6.4e" % (chi2 / float(N - len(pfit) + 1)), end=" ")
        if follow is None:
            print("")
        else:
            try:
                print(" ".join([k + "=" + "%5.2e" % params[k] for k in follow]))
            except Exception:
                print("")
    return res
